<?php

namespace Drupal\field_redirect\Service;

use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Component\Render\MarkupInterface;
use Drupal\Core\Config\Config;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\ContentEntityFormInterface;
use Drupal\Core\Entity\EditorialContentEntityBase;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\Render\Markup;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\field\FieldStorageConfigInterface;
use Drupal\file\FileInterface;
use Drupal\user\UserInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpKernel\Exception\HttpException;

/**
 * Utility service class.
 *
 * phpcs:disable Drupal.Semantics.FunctionT.NotLiteralString
 */
class UtilityService {

  use StringTranslationTrait;
  use MessengerTrait;

  public const MODULE_NAME = 'field_redirect';
  private const DEFAULT_REDIRECT_STATUS = 307;
  private const ENTITY_PAIR_SEPARATOR = ':';
  private const ENTITY_BUNDLE_WILDCARD = '*';
  private const PAIR_SEPARATOR = '|';
  private const FIELDS_SEPARATOR = ',';
  private const STATUS_SEPARATOR = '#';
  private const CR = "\r";
  private const LF = "\n";
  private const CRLF = self::CR . self::LF;
  private const SUPPORTED_FIELD_TYPES = [
    'link',
    'file',
    'image',
  ];

  /**
   * The module settings.
   */
  private Config $settings;

  /**
   * The request service.
   */
  private Request $request;

  /**
   * The error message assoc.
   */
  private array $errors = [];

  /**
   * The line number.
   */
  private int $lineNumber = 0;

  /**
   * Service constructor.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    RequestStack $request_stack,
    private EntityTypeManagerInterface $entityTypeManager,
    private EntityFieldManagerInterface $entityFieldManager,
    private FileUrlGeneratorInterface $fileUrlGenerator,
  ) {
    $this->settings = $config_factory->get(self::MODULE_NAME . '.settings');
    $this->request = $request_stack->getCurrentRequest();
  }

  /**
   * Convert the mappings array to a string.
   */
  public function toString(array $mappings_array = []): string {
    $mappings_string = '';
    foreach ($mappings_array as $entity_type_id__bundle => $fields) {
      if (\str_starts_with($entity_type_id__bundle, 'user' . self::ENTITY_PAIR_SEPARATOR)) {
        $entity_type_id__bundle = 'user';
      }
      $mappings_string .= $this->desorptEntityAndFields('implode', [
        $entity_type_id__bundle,
        $this->desorptFields('implode', $fields),
      ]) . self::LF;
    }
    return (string) $this->trim($mappings_string);
  }

  /**
   * Convert a mappings string to an array.
   */
  public function toArray(string $mappings_string = ''): array {
    $mappings_string = $this->trim($mappings_string);
    if ($mappings_string === '') {
      return [];
    }
    $mappings_array = [];
    if ($lines = $this->trim(\explode(self::LF, $mappings_string))) {
      foreach ($lines as $entity_type_id__bundle__fields) {
        [$entity_type_id__bundle, $fields] = $this->desorptEntityAndFields('explode', $entity_type_id__bundle__fields);
        $mappings_array[$entity_type_id__bundle] = \array_map(fn ($field_machine_name__status) => $this->desorptFieldMachineNameAndStatus('explode', $field_machine_name__status), $this->desorptFields('explode', $fields));
      }
    }
    return $mappings_array;
  }

  /**
   * Smart trimmer.
   */
  public function trim(string|array $string): string|array {
    switch (\gettype($string)) {
      case 'object':
        $string = (array) $string;
      case 'array':
        foreach ($string as $k => $v) {
          $string[$k] = $this->trim($v);
        }
        return $string;

      case 'string':
        $string = \str_replace(self::CRLF, self::LF, $string);
        $string = \str_replace(self::CR, self::LF, $string);
        $string = \preg_replace('/\n+/', self::LF, $string);
        return \mb_ereg_replace('^[[:space:]]*([\s\S]*?)[[:space:]]*$', '\1', $string);

      default:
        throw new \Exception('Specified value type is invalid.');
    }
  }

  /**
   * Validate that the mappings is correct.
   */
  public function isValidMappings(FormStateInterface $form_state, array $element, string $mappings_string = ''): bool {
    $mappings_string = $this->trim($mappings_string);
    if ($mappings_string === '') {
      return TRUE;
    }
    if ($lines = $this->trim(\explode(self::LF, $mappings_string))) {
      $this->errors = [];
      $this->lineNumber = 0;
      $duplicates_entity = [];
      foreach ($lines as $entity_type_id__bundle__fields) {
        if ($entity_type_id__bundle__fields === '') {
          continue;
        }
        ++$this->lineNumber;
        [$entity_type_id__bundle, $fields] = $this->desorptEntityAndFields('explode', $entity_type_id__bundle__fields);
        [$entity_type_id, $bundle] = $this->desorptTypeAndBundle('explode', $entity_type_id__bundle);
        if ($entity_type_id === '' || ($entity_type_id !== 'user' && $bundle === '') || $fields === '') {
          $this->mergeErrorMessage('This mapping format is incorrect.');
          continue;
        }
        $duplicates_entity[$entity_type_id__bundle] ??= 0;
        ++$duplicates_entity[$entity_type_id__bundle];
        $field_storage_definitions = $this->entityFieldManager->getFieldStorageDefinitions($entity_type_id);
        if (empty($field_storage_definitions)) {
          $this->mergeErrorMessage('The bundle name "@bundle" does not exist for entity type "@entity_type_id".', [
            '@bundle' => $bundle,
            '@entity_type_id' => $entity_type_id,
          ]);
        }
        $field_machine_names = $this->desorptFields('explode', $fields);
        $duplicates_field = [];
        $max_i = \count($field_machine_names) - 1;
        foreach ($field_machine_names as $current_i => $field_machine_name__status) {
          $duplicates_field[$field_machine_name__status] ??= 0;
          ++$duplicates_field[$field_machine_name__status];
          [$field_machine_name, $status] = $this->desorptFieldMachineNameAndStatus('explode', $field_machine_name__status);
          if (!empty($field_machine_name)) {
            $field_type = NULL;
            if ($field_storage_definition = $field_storage_definitions[$field_machine_name] ?? FALSE) {
              if ($field_storage_definition instanceof BaseFieldDefinition
                || ($field_storage_definition instanceof FieldStorageConfigInterface
                  && ($bundle === self::ENTITY_BUNDLE_WILDCARD
                    || ($field_storage_definition->getBundles()[$bundle] ?? FALSE)
                  )
                )
              ) {
                $field_type = $field_storage_definition->getType();
              }
            }
            if (!isset($field_type)) {
              if ($bundle === self::ENTITY_BUNDLE_WILDCARD) {
                $this->mergeErrorMessage('The field for "@field_machine_name" does not exist for entity type "@entity_type_id".', [
                  '@field_machine_name' => $field_machine_name,
                  '@entity_type_id' => $entity_type_id,
                ]);
              }
              else {
                $this->mergeErrorMessage('The field for "@field_machine_name" does not exist for bundle "@bundle" of entity type "@entity_type_id".', [
                  '@field_machine_name' => $field_machine_name,
                  '@bundle' => $bundle,
                  '@entity_type_id' => $entity_type_id,
                ]);
              }
            }
            elseif (!\in_array($field_type, self::SUPPORTED_FIELD_TYPES, TRUE)) {
              $this->mergeErrorMessage('The field type for "@field_machine_name" must be link, file or image.', [
                '@field_machine_name' => $field_machine_name,
              ]);
            }
          }
          if (!empty($status)) {
            if (\str_starts_with($status, '3')) {
              if ($field_machine_name === '') {
                $this->mergeErrorMessage('The field should be specified if status "#3xx" is specified.');
              }
              elseif (!\in_array($status, ['301', '302', '303', '307'], TRUE)) {
                $this->mergeErrorMessage('The specified status is invalid. Please specify "#301", "#302", "#303" or "#307".');
              }
            }
            if (\str_starts_with($status, '4')) {
              if ($field_machine_name !== '') {
                $this->mergeErrorMessage('The field cannot be specified if status "#4xx" is specified.');
              }
              elseif (!\in_array($status, ['403', '404'], TRUE)) {
                $this->mergeErrorMessage('The specified status is invalid. Please specify "#403" or "#404".');
              }
              if ($max_i !== $current_i) {
                $this->mergeErrorMessage('Status is can be specified only one at the end.');
              }
            }
          }
        }
        foreach ($duplicates_field as $field => $count) {
          if ($count > 1) {
            if (\str_starts_with($field, '#4')) {
              $this->mergeErrorMessage('Multiple specification of status "#4xx" is not allowed.');
            }
            else {
              $this->mergeErrorMessage('Fields for "@field_machine_name" are specified is duplicate.', [
                '@field_machine_name' => $field_machine_name,
              ]);
            }
          }
        }
      }
    }
    foreach ($duplicates_entity as $entity_type_bundle => $count) {
      if ($count > 1) {
        $this->mergeErrorMessage('Set of the entity type and bundle is duplicated. (@entity_type_bundle)', [
          '@entity_type_bundle' => $entity_type_bundle,
        ], FALSE);
      }
    }
    if (!empty($this->errors)) {
      $form_state->setError($element, Markup::create(\implode('<br>', \array_keys($this->errors))));
      return FALSE;
    }
    return TRUE;
  }

  /**
   * Form alter.
   */
  public function formAlter(&$form, FormStateInterface $form_state, $form_id): void {
    $form_object = $form_state->getFormObject();
    if ($form_object instanceof ContentEntityFormInterface) {
      $entity = $form_object->getEntity();
      if ($entity instanceof EntityInterface) {
        $redirect_fields = $this->getRedirectFields($entity->getEntityTypeId(), $entity->bundle());
        foreach ($redirect_fields as $field_machine_name__status) {
          [$field_machine_name, $status] = $field_machine_name__status;
          if ($widget = &$form[$field_machine_name]['widget'][0] ?? FALSE) {
            $message = $this->t('<p><strong>Field Redirect Module: If this field value is valid, redirect to the field value URL with redirect status "@status".</strong></p>', [
              '@status' => $this->getHttpStatusMessage($status ?: self::DEFAULT_REDIRECT_STATUS),
            ]);
            if (\in_array(($widget['uri']['#type'] ?? ''), ['url', 'entity_autocomplete'], TRUE)) {
              $description = $widget['uri']['#description'] ?? '';
              $widget['uri']['#description'] = $message . $description;
            }
            elseif (($widget['#type'] ?? '') === 'managed_file') {
              $description = $widget['#description'] ?? '';
              $widget['#description'] = $message . $description;
            }
          }
        }
      }
    }
  }

  /**
   * Redirect to the specified destination or return the status.
   */
  public function doRedirect(string $mode, EntityInterface $entity): void {
    if ($entity = $this->checkEntityInstance($entity, TRUE)) {
      $target_entity_type = $entity->getEntityTypeId();
      $target_entity_bundle = $entity->bundle();
      $target_entity_id = $entity->id();

      // Check exists target field in entities bundle.
      $redirect_fields = $this->getRedirectFields($target_entity_type, $target_entity_bundle);

      $field_machine_name = '';
      $field_value = [];
      $status = '';
      foreach ($redirect_fields as $field_machine_name__status) {
        [$_field_machine_name, $status] = $field_machine_name__status;
        if ($entity->hasField($_field_machine_name) && ($_field_value = $entity->get($_field_machine_name)->getValue())) {
          $field_machine_name = $_field_machine_name;
          $field_value = $_field_value;
          break;
        }
      }
      if ($field_machine_name === '' && $status === '') {
        return;
      }

      // The "taxonomy_term" path seems to be a little different
      // from the others.
      $target_entity_type_path = ($target_entity_type === 'taxonomy_term')
        ? 'taxonomy/term'
        : $target_entity_type;

      switch ($mode) {
        case 'after_save':
          // If exist "destination" in current url parameters then
          // this function process ending here, because to give
          // priority to "destination".
          if ($this->request->query->get('destination')) {
            return;
          }

          // Get label of entities bundle.
          $target_bundle_label = $this->getEntityTypeLabel($entity);

          // 4xx.
          if (\str_starts_with($status, '4')) {
            // Display a status message descriptioning this
            // behavior to avoid editor confusion.
            $this->messenger()->addStatus($this->t('Field Redirect Module: The original behavior was that if you accessed this %target_bundle_label view directly, then return status of the @status, but because I had just edited it, was redirected to the %target_bundle_label edit form.', [
              '%target_bundle_label' => $target_bundle_label,
              '@status' => $this->getHttpStatusMessage($status),
            ]));
          }

          // Redirect.
          else {
            // Display a status message descriptioning this
            // behavior to avoid editor confusion.
            $this->messenger()->addStatus($this->t('Field Redirect Module: The original behavior was that if you accessed this %target_bundle_label view directly, would be redirect to %a_tag (@status), but since it was just edited, was redirected to the %target_bundle_label edit form.', [
              '%target_bundle_label' => $target_bundle_label,
              '%a_tag' => $this->getUrl($field_value, TRUE),
              '@status' => $this->getHttpStatusMessage($status),
            ]));
          }

          // Return to the edit form.
          $redirect_path = '/' . $target_entity_type_path . '/' . $target_entity_id . '/edit';
          $redirect_url = Url::fromUserInput($redirect_path)->toString();
          $redirect_response = new RedirectResponse($redirect_url);
          $redirect_response->send();
          break;

        case 'view':
          // 4xx.
          if (\str_starts_with($status, '4')) {
            throw new HttpException($status);
          }

          // Redirect.
          elseif ($redirect_url = $this->getUrl($field_value)) {
            // Avoid redirect loops.
            $site_url = $this->request->getSchemeAndHttpHost();
            $site_url__no_scheme = \preg_replace('/^[^:]+/', '', $site_url);
            $current_uri = $this->request->getRequestUri();
            $current_url__http = 'http' . $site_url__no_scheme . $current_uri;
            $current_url__https = 'https' . $site_url__no_scheme . $current_uri;
            $current_entity_uri = '/' . $target_entity_type_path . '/' . $target_entity_id;
            $current_entity_url__http = 'http' . $site_url__no_scheme . $current_entity_uri;
            $current_entity_url__https = 'https' . $site_url__no_scheme . $current_entity_uri;
            if (FALSE
              || $redirect_url === $current_uri
              || $redirect_url === $current_url__http
              || $redirect_url === $current_url__https
              || $redirect_url === $current_entity_uri
              || $redirect_url === $current_entity_url__http
              || $redirect_url === $current_entity_url__https
            ) {
              return;
            }

            // Do redirect to $redirect_url.
            (new RedirectResponse($redirect_url, $status ?: self::DEFAULT_REDIRECT_STATUS))
              ->send();
            exit;
          }
          break;
      }
    }
  }

  /**
   * Add error message without duplication.
   */
  private function mergeErrorMessage(string $message, array $args = [], bool $lineNumber = TRUE): void {
    if ($lineNumber) {
      $message = $this->t('Line @lineNumber:')->__toString() . ' ' . $message;
      $args += ['@lineNumber' => $this->lineNumber];
    }
    $error_key = (new TranslatableMarkup($message, $args))->__toString();
    $this->errors[$error_key] = 1;
  }

  /**
   * Explode and Implode the "entity type + entity bundle" & "fields" pair.
   */
  private function desorptEntityAndFields(string $mode, string|array $entity_type_id__bundle__fields): array|string {
    switch ($mode) {
      case 'explode':
        $pair = $this->trim(\explode(self::PAIR_SEPARATOR, $entity_type_id__bundle__fields, 2));
        $entity_type_id__bundle = $pair[0] ?? '';
        $fields = $pair[1] ?? '';
        return [$entity_type_id__bundle, $fields];

      case 'implode':
        $entity_type_id__bundle__fields = (array) $this->trim($entity_type_id__bundle__fields);
        if (empty($entity_type_id__bundle__fields)) {
          return '';
        }
        if (empty($entity_type_id__bundle__fields[1])) {
          return (string) $entity_type_id__bundle__fields[0];
        }
        return \implode(self::PAIR_SEPARATOR, $entity_type_id__bundle__fields);

      default:
        throw new \Exception('Specified $mode the "' . $mode . '" is invalid.');
    }
  }

  /**
   * Explode and Implode the "entity type" & "entity bundle" pair.
   */
  private function desorptTypeAndBundle(string $mode, string|array $entity_type_id__bundle): array|string {
    switch ($mode) {
      case 'explode':
        $pair = $this->trim(\explode(self::ENTITY_PAIR_SEPARATOR, $entity_type_id__bundle, 2));
        $entity_type_id = $pair[0] ?? '';
        $bundle = $pair[1] ?? '';
        return [$entity_type_id, $bundle];

      case 'implode':
        $entity_type_id__bundle = (array) $this->trim($entity_type_id__bundle);
        if (empty($entity_type_id__bundle)) {
          return '';
        }
        if (empty($entity_type_id__bundle[1])) {
          return (string) $entity_type_id__bundle[0];
        }
        return \implode(self::ENTITY_PAIR_SEPARATOR, $entity_type_id__bundle);

      default:
        throw new \Exception('Specified $mode the "' . $mode . '" is invalid.');
    }
  }

  /**
   * Explode and Implode the "fields".
   */
  private function desorptFields(string $mode, string|array $fields): array|string {
    switch ($mode) {
      case 'explode':
        return $this->trim(\explode(self::FIELDS_SEPARATOR, $fields));

      case 'implode':
        return \implode(self::FIELDS_SEPARATOR, \array_map(fn ($field_machine_name__status) => $this->desorptFieldMachineNameAndStatus('implode', $field_machine_name__status), (array) $fields));

      default:
        throw new \Exception('Specified $mode the "' . $mode . '" is invalid.');
    }
  }

  /**
   * Explode and Implode the "field_machine_name" & "status" pair.
   */
  private function desorptFieldMachineNameAndStatus(string $mode, string|array $field_machine_name__status): array|string {
    switch ($mode) {
      case 'explode':
        $pair = $this->trim(\explode(self::STATUS_SEPARATOR, (string) $field_machine_name__status, 2));
        $field_machine_name = $pair[0] ?? '';
        $status = $pair[1] ?? '';
        return [$field_machine_name, $status];

      case 'implode':
        $field_machine_name__status = (array) $this->trim($field_machine_name__status);
        if (empty($field_machine_name__status)) {
          return '';
        }
        if (empty($field_machine_name__status[1])) {
          return (string) $field_machine_name__status[0];
        }
        return \implode(self::STATUS_SEPARATOR, $field_machine_name__status);

      default:
        throw new \Exception('Specified $mode the "' . $mode . '" is invalid.');
    }
  }

  /**
   * Get redirect fields the specified bundle.
   */
  private function getRedirectFields(string $target_entity_type, string $target_entity_bundle): array {
    $mappings = $this->settings->get(self::MODULE_NAME . '.mappings');
    $wildcard_entity_type = $this->desorptTypeAndBundle('implode', [
      $target_entity_type,
      self::ENTITY_BUNDLE_WILDCARD,
    ]);
    $specified_entity_type = $this->desorptTypeAndBundle('implode', [
      $target_entity_type,
      $target_entity_bundle,
    ]);
    $redirect_fields = $mappings[$wildcard_entity_type] ?? [];
    $redirect_fields += $mappings[$specified_entity_type] ?? [];
    return $redirect_fields;
  }

  /**
   * Get label from specified entity.
   */
  private function getEntityTypeLabel(EntityInterface $entity): string {
    return $entity->getEntityType()->get('label')?->render() ?? '[unknown]';
  }

  /**
   * Get html link tag markup.
   */
  private function getUrl(array $target_field_value, bool $need_link_tag = FALSE): MarkupInterface|string|null {
    // If exist value of URL.
    if ($uri = $target_field_value[0]['uri'] ?? FALSE) {
      // Generate of url from $uri.
      $url = Url::fromUri($uri, ['attributes' => ['target' => '_blank']]);

      return $need_link_tag
        ? Link::fromTextAndUrl($url->toString(), $url)->toString()
        : $url->toString();
    }

    // If exist value of target id of th file entity.
    elseif ($target_id = $target_field_value[0]['target_id'] ?? FALSE) {
      if (($file = $this->entityTypeManager->getStorage('file')->load($target_id)) instanceof FileInterface) {
        if ($uri = $file->getFileUri()) {
          // File URL.
          $url = $this->fileUrlGenerator->generateAbsoluteString($uri);

          // Link text.
          $text = $file->getFilename();
          if (($description = (string) ($target_field_value[0]['description'] ?? '')) !== '') {
            $text .= ' (' . $description . ')';
          }

          return $need_link_tag
            ? Markup::create('<a href="' . $url . '" target="_blank">' . $text . '</a>')
            : $url;
        }
      }
    }
    return NULL;
  }

  /**
   * Validate if an entity is allowed and return that entity.
   */
  private function checkEntityInstance(EntityInterface $entity, bool $check_status = FALSE): EditorialContentEntityBase|UserInterface|false {
    if ($entity instanceof EditorialContentEntityBase) {
      if ($check_status) {
        return $entity->isPublished() ? $entity : FALSE;
      }
      return $entity;
    }
    if ($entity instanceof UserInterface) {
      if ($check_status) {
        return $entity->isActive() ? $entity : FALSE;
      }
      return $entity;
    }
    return FALSE;
  }

  /**
   * Get the HTTP status message from the HTTP status code.
   */
  private function getHttpStatusMessage(string $status_code): string {
    return match ($status_code) {
      '403' => '403 Forbidden',
      '404' => '404 Not Found',
      '301' => '301 Moved Permanently',
      '302' => '302 Found',
      '303' => '303 See Other',
      '307' => '307 Temporary Redirect',
      default => '',
    };
  }

}
